# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""FletsAzukeru File Access module.

Define classes for FletsAzukeru file access.
"""

# common module
from __future__ import unicode_literals
from fileaccess import (FileObj, Access)
import os
import stat
import arrow
from buffer_object import BufferObject, BufferObjectW, ChunkedBuffer
from fileaccess_helper import *

# feature module
import xml.etree.ElementTree as ElementTree
from pathid import PathIdMap
from util import UploadFileData
import random
import string
import httplib2
import urllib
import re
import proxy_info
import threading
import requests
from time import sleep
import sys
import json

# common constant
FILE_SEPARATOR = "/"
FILE_MODE = 0
DIR_MODE = stat.S_IFDIR
ERR_MSG_RENAME = "Invalid new name"
ERR_MSG_INVALID_FILE_LENGTH = "Invalid file length"
ERR_MSG_INVALID_OPEN_MODE = "Invalid open mode"
OPEN_MODE_READ = "r"
OPEN_MODE_WRITE = "w"

# feature constant
CONTENT_TYPE__URL_ENCODED = "application/x-www-form-urlencoded"
CONTENT_TYPE__FORM_DATA = "multipart/form-data"
KEY_COMPANY_ID = "X-FBS-Company-id"
KEY_SERVICE_ID = "X-FBS-Service-id"
KEY_SESSION_ID = "X-FBS-Session-id"
KEY_TERMINAL_AUTH_ID = "X-FBS-Terminal-auth-id"
ROOT_DIR_NAME = {
    u"my": u"マイフォルダ",
    u"group": u"グループフォルダ",
    u"backup": u"自動バックアップフォルダ",
}


class FletsAzukeruFileObj(FileObj):
    """Class for FletsAzukeru file."""

    def __init__(self, access, path, mode, length, chunk_size):
        """Constructor."""

        super(FletsAzukeruFileObj, self).__init__()
        self._access = access
        self._path = path
        self._mode = (filter((lambda m: m in mode.lower()), (OPEN_MODE_READ, OPEN_MODE_WRITE)) or [None])[0]
        self._length = length
        self._chunk_size = chunk_size
        if self._chunk_size is None:
            self._chunk_size = 8 * 1024
        self._r_pos = 0
        self._w_pos = 0
        self._downloader = None
        self._uploader = None
        self._uploader_exception = None
        self._id = ""

        if self._mode == OPEN_MODE_READ:
            self._length = self._access.stat_file(self._path)[stat.ST_SIZE]
            self._download()
        elif self._mode == OPEN_MODE_WRITE:
            if self._length < 0:
                raise ValueError(ERR_MSG_INVALID_FILE_LENGTH)
            self._upload()
        else:
            raise ValueError(ERR_MSG_INVALID_OPEN_MODE)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        """Close file."""

        try:
            if self._mode == OPEN_MODE_READ:
                pass
            elif self._mode == OPEN_MODE_WRITE:
                if self._uploader:
                    self._uploader.join()
        finally:
            if self._buffer:
                self._buffer.close()
            if self._uploader_exception:
                uploader_exception = self._uploader_exception
                self._uploader_exception = None
                raise uploader_exception[1], None, uploader_exception[2]

            if self._id:
                self._access.path_id_map.append_path(self._path, self._id)

    def _download(self):
        """Download to buffer"""

        if self._downloader is None:
            self._buffer = BufferObject(size=self._chunk_size)
            item_id = self._access._get_item_id(self._path, be_raise=(self._mode == OPEN_MODE_READ))
            if item_id is not None:
                _headers, _url = self._access._gen_request_headers_url(
                    uri=["file", item_id, "download"],
                )
                r = requests.get(url=_url, headers=_headers, stream=True,
                        verify=False, proxies={"https": self._access._proxy},
			timeout=300)
                self._downloader = r.iter_content(chunk_size=self._chunk_size)

        self._buffer.set_w_mode()
        self._buffer.clear()
        while self._buffer.lack > 0:
            try:
                data = self._downloader.next()
                self._buffer.write(data)
            except StopIteration:
                break
        self._buffer.set_r_mode()

    def _upload(self):
        if self._uploader is None:
            self._buffer = ChunkedBuffer(
                size=self._length,
                buffer_size=self._chunk_size,
                read_at_first=True,
                retry_limit=30000,
                delay=0.01
            )
            self._uploader = threading.Thread(target=self._upload_proc)
            self._uploader.start()

    def _upload_proc(self):
        """Upload from buffer"""

        name = os.path.basename(self._path)
        parent = os.path.dirname(self._path)
        parent_id = self._access._get_item_id(parent)

        boundary = "boundary_%s" % (''.join(
            [random.choice(string.ascii_letters + string.digits) for i in range(25)]))

        CustomUploadFileData = UploadFileData
        CustomUploadFileData.FILE_NAME = "upload_file".encode("utf-8")

        self._buffer.set_r_mode()
        ufd = CustomUploadFileData(name.encode("utf-8"), self._buffer, -1, boundary.encode("utf-8"))
        headers = ufd.get_headers()

        try:
            resp = self._access._request(
                method="POST",
                uri=["folder", parent_id, "upload"],
                headers=headers,
                body=ufd,
            )
            try:
                self._id = resp["body"].findtext("file_id")
            except:
                self._id = ""

        except Exception as e:
            self._uploader_exception = sys.exc_info()
            self._buffer.close()

    def read(self, size=-1):
        """Return a "size" bytes data from current position in file.
        If the size argument is negative or omitted, then return all data.
        """

        if size < 0:
            read_size = self._length
        else:
            read_size = min(size, self._length - self._r_pos)

        ret_buf = BufferObjectW(size=read_size)
        while ret_buf.lack > 0:
            ret_buf.write(self._buffer.read(ret_buf.lack))
            if self._buffer.rest == 0:
                self._buffer.close()
                self._download()

        self._r_pos += ret_buf.length
        ret = ret_buf.read()
        ret_buf.close()

        return ret

    def write(self, data):
        """Write a "data" data to current position in file."""

        if self._uploader.is_alive():
            self._buffer.set_w_mode()
            self._buffer.write(data)
            self._buffer.close_chunk()


class FletsAzukeruAccess(Access):
    """class accessing to FletsAzukeru file."""

    def __init__(self,
                 client_id=None,
                 client_secret=None,
                 host_name=None,
                 company_id=None,
                 service_id=None,
                 root_dir=None,
                 proxy=None,
                 is_secure=True,
                 terminal_id=None,
                 terminal_key=None,
                 using_image_login=False,
                 conf_path="",
                 retry_limit=10,
                 retry_delay=10.0
                 ):
        """Constructor.

        Args:
            client_id (string): client identifier.
            client_secret (string): client secret.
            host_name (string): host name.
            company_id (string): company id.
            service_id (string): service id.
            root_dir (string): Root Folder
                my: マイフォルダ
                group: グループフォルダ
                backup: 自動バックアップフォルダ
            proxy (string): proxy setting
                If None, bypass proxies
                e.g.
                proxy = 'http://username:password@127.0.0.1:8888'
            is_secure (bool):
                if true(Default), URI schema is 'https'
                else, 'http'
            retry_limit (int):
                The maximum number of retries
                when getting "320 Found" response while trying to log in
            retry_delay (float):
                delay time to retry
                when getting "320 Found" response while trying to log in
        """

        super(FletsAzukeruAccess, self).__init__()

        self._client_id = client_id
        self._client_secret = client_secret
        self._host_name = host_name
        self._company_id = company_id
        self._service_id = service_id
        self._proxy = proxy
        self._root_dir = root_dir
        self._is_secure = is_secure
        self._using_image_login = using_image_login
        self._conf_path = conf_path
        self._terminal_id = terminal_id
        self._terminal_key = terminal_key
        self._retry_limit = retry_limit
        self._retry_delay = retry_delay

        self.path_id_map = PathIdMap(separator=FILE_SEPARATOR)

        if ROOT_DIR_NAME.get(self._root_dir) is None:
            raise Exception("Unknown Root Folder")

        self._session_id = None
        self._terminal_auth_id = None

        headers = {
            'content-type': CONTENT_TYPE__URL_ENCODED,
        }
        data = {
            'login_id': self._client_id,
            'password': self._client_secret,
        }

        login_uri = 'login'
        if self._using_image_login:
            login_uri = 'image_auth/login'
        resp = self._request(
            method="POST",
            uri=login_uri,
            headers=headers,
            body=urllib.urlencode(data),
        )

        self._session_id = resp["header"]["x-fbs-session-id"]
        self._terminal_auth_id = self._try_get_terminal_auth_id(headers)

    def __del__(self):
        """Destructor."""

        if self._session_id is not None:
            self._request(
                    method="DELETE",
                    uri="login",
                )

    @property
    def auth_url(self):
        return None

    def authenticate(self, params=None):
        pass

    def _gen_request_headers_url(self, uri=None, query=None, headers=None):
        if isinstance(uri, str):
            uri = [uri]
        elif isinstance(uri, unicode):
            uri = [uri]

        if uri[0] != "service":
            uri.insert(0, "service")
        if uri[1] != "fbs":
            uri.insert(1, "fbs")

        url = "http%s://%s/%s" % (
            "s" if self._is_secure else "", self._host_name, "/".join(uri))
        if query is not None:
            url = url + "?" + urllib.urlencode(query, True)

        if headers is None:
            headers = {}

        if headers.get(KEY_COMPANY_ID) is None:
            headers[KEY_COMPANY_ID] = self._company_id

        if headers.get(KEY_SERVICE_ID) is None:
            headers[KEY_SERVICE_ID] = self._service_id

        if headers.get(KEY_SESSION_ID) is None and self._session_id is not None:
            headers[KEY_SESSION_ID] = self._session_id

        if headers.get(KEY_TERMINAL_AUTH_ID) is None and self._terminal_auth_id is not None:
            headers[KEY_TERMINAL_AUTH_ID] = self._terminal_auth_id

        for key in headers:
            if not isinstance(headers[key], unicode):
                headers[key] = unicode(headers[key])

        return headers, url

    def _request(self, method=None, uri=None, query=None, headers=None, body=None):
        headers, url = self._gen_request_headers_url(uri, query, headers)

        if self._proxy is None:
            h = httplib2.Http(timeout=300, disable_ssl_certificate_validation=True)
        else:
            h = httplib2.Http(timeout=300,
                proxy_info=proxy_info.proxy_info_from_url(self._proxy),
                disable_ssl_certificate_validation=True
            )

        cnt = 0
        while True:
            resp, content = h.request(
                url,
                method=method,
                headers=headers,
                body=body,
            )
            code = resp.status
            reason = resp.reason
            error_code = resp.get('x-fbs-error-code')
            error_string = resp.get('x-fbs-error-string')
#            if isinstance(content, str):
#                print content
#            elif isinstance(content, unicode):
#                print content.encode("utf-8")

            if code in [200, 201]:
                break
            elif cnt >= self._retry_limit:
                raise httplib2.HttpLib2Error(
                    "Retry count exceed the limit/"
                    "code:'%d' reason:'%s' error_code:%s error_string:'%s' (retry count: %d)" % (
                        code, reason, error_code, error_string, cnt))
            elif code in [302]:
                # retry
                cnt += 1
                sleep(self._retry_delay)
            else:
                raise httplib2.HttpLib2Error(
                    "code:'%d' reason:'%s' error_code:%s error_string:'%s'" % (code, reason, error_code, error_string))

        xml = None
        try:
            _content_type = resp.get("content-type")
            content_types = _content_type.split(";")
            content_type = content_types[0]
            if content_type == "text/xml":
                xml = ElementTree.fromstring(content)
        except:
            pass

        ret = {
            "header": resp,
            "body": xml,
            "body_raw": content,
        }

        return ret

    def _stat(self, target_path=None, target_item=None):
        """Return a stat information dictionary for "target_path"."""

        if target_item is None:
            item_id = self._get_item_id(target_path, be_raise=True)
            resp = self._request(
                method="GET",
                uri=["file", item_id],
            )
            target_item = resp["body"].find("file")

        _mode = FILE_MODE
        folder_flg = target_item.findtext("folder_flg")
        if folder_flg is not None:
            if folder_flg == "1":
                _mode = DIR_MODE
        elif target_item.findtext("folder_name") is not None:
            _mode = DIR_MODE

        _mtime = target_item.findtext("modified_time")
        if _mtime is not None:
            _mtime = arrow.get(_mtime, "YYYY/MM/DD HH:mm:ss").replace(tzinfo="+09:00").timestamp

        _ctime = None
        _size = None
        if _mode == FILE_MODE:
            _size = target_item.findtext("file_size")
            if _size is not None:
                _size = int(_size)

        return {
            stat.ST_MODE: _mode,
            stat.ST_MTIME: _mtime,
            stat.ST_CTIME: _ctime,
            stat.ST_SIZE: _size,
        }

    def _get_root_id(self):
        resp = self._request(
            method="GET",
            uri=["folder", "list"]
        )

        matches = []
        xml = resp["body"]
        target_path = u"//%s" % (ROOT_DIR_NAME.get(self._root_dir))
        items = xml.findall("folders/folder")
        for item in items:
            file_id = item.findtext("file_id")
            file_path = item.findtext("file_path")
            if file_path == target_path:
                matches.append(file_id)

        if len(matches) == 1:
            return matches[0]
        elif len(matches) == 0:
            return None
        else:
            raise Exception("multiple items with same name")

    def _get_directory(self, path, be_raise=False):
        return self._get_item_id(path, be_raise)

    def _get_file(self, path, be_raise=False):
        return self._get_item_id(path, be_raise)

    def _debug_out(self, msg):

#        if isinstance(msg, unicode):
#            msg = msg.encode("UTF8")
#        open("/dev/pts/1", "w").write(msg)
        pass

    def _get_item_id(self, path, be_raise=False):
        self._debug_out("_get_item_id called for %s\n" % (path))
        if path == "":
            path = FILE_SEPARATOR
        try:
            item_id = self.path_id_map.get_id(path)
            if item_id != None:
                self._debug_out("Found id %s - %s\n" % (path, item_id))
                return item_id
        except:
            pass

        if path == FILE_SEPARATOR:
            item_id = self._get_root_id()
            self.path_id_map.append_path(FILE_SEPARATOR, item_id)
            self._debug_out("New root id %s - %s\n" % (path, item_id))
            return item_id

        parent_path = FILE_SEPARATOR
        sep_pos = path.rfind(FILE_SEPARATOR)
        if sep_pos > 0:
            parent_path = path[:sep_pos]
        try:
            parent_id = self.path_id_map.get_id(parent_path)
            self._debug_out("Found parent %s - %s\n" % (parent_path, parent_id))
        except:
            parent_id = None

        if parent_id == None:
            parent_id = self._get_item_id(parent_path, be_raise)
            self._debug_out("Get parent id %s - %s\n" % (parent_path, parent_id))
            if parent_id != None:
                self._debug_out("Append parent id %s - %s\n" % (parent_path, parent_id))
                self.path_id_map.append_directory(parent_path, parent_id)

        if parent_id == None:
            if be_raise:
                raise LookupError(path)
            else:
                return None

        found_id = None
        for item in self._get_item_list(parent_path=parent_path,
                                            parent_id=parent_id):

            item_name = os_path_bottom_name(item.findtext("file_path"))
            item_id = item.findtext("file_id")
            if parent_path and parent_path[-1] == FILE_SEPARATOR:
                item_path = parent_path + item_name
            else:
                item_path = parent_path + FILE_SEPARATOR + item_name

            self.path_id_map.append_path(item_path, item_id)
            self._debug_out("Append item id %s - %s\n" % (item_path, item_id))

            if item_path == path:
                found_id = item_id

        if found_id == None:
            if be_raise:
                raise LookupError(path)

        return found_id

    def _get_item_list(self, parent_path="", parent_id=None):
        ret = []
        if parent_id is None:
                parent_id = self._get_item_id(parent_path, be_raise=True)

        idx = 0
        limit =300
        while True:
            query = {"limit": limit, "offset": idx * limit}
            resp = self._request(
                method="GET",
                query=query,
                uri=["folder", parent_id, "list", ],
                body = {}
            )
            files = resp["body"].findall("files/file")
            folders = resp["body"].findall("folders/folder")
            ret.extend(files)
            ret.extend(folders)
            if len(files) + len(folders) == 0:
                break
            else:
                idx += 1

        return ret

    def _delete_item(self, target_path=None, target_id=None):
        if target_id is None:
            target_id = self._get_item_id(target_path, be_raise=True)

        body = urllib.urlencode({"file_id[]": [target_id.encode("utf-8")]}, True)

        try:
            self._request(
                method="POST",
                uri=["file", "delete"],
                headers={"content-type": CONTENT_TYPE__URL_ENCODED},
                body=body,
            )
        except httplib2.HttpLib2Error, exc_ins:
            if str(exc_ins).find("error_code:13 ") >= 0:
                pass
            else:
                raise

        try:
            self.path_id_map.delete_path(target_path)
        except:
            pass

    def _rename_item(self, target_path=None, target_id=None, new_name=None):
        if FILE_SEPARATOR in new_name:
            raise ERR_MSG_RENAME

        if target_id is None:
            target_id = self._get_item_id(target_path, be_raise=True)

        self._request(
            method="PUT",
            uri=["file", target_id],
            headers={"content-type": CONTENT_TYPE__URL_ENCODED},
            body=urllib.urlencode({"file_name": new_name.encode("utf-8")})
        )

        try:
            self.path_id_map.delete_path(target_path)
        except:
            pass
        try:
            self.path_id_map.append_path(target_path, target_id)
        except:
            pass

    def _move_item(self, target_path=None, target_id=None, new_dir_path=None):
        if target_id is None:
            target_id = self._get_item_id(target_path, be_raise=True)

        new_dir_id = self._get_item_id(new_dir_path, be_raise=True)

        self._request(
            method="POST",
            uri=["folder", new_dir_id, "move_to"],
            headers={"content-type": CONTENT_TYPE__URL_ENCODED},
            body=urllib.urlencode({"file_id[]": [target_id.encode("utf-8")]}, True)
        )

        try:
            self.path_id_map.delete_path(target_path)
        except:
            pass
        try:
            self.path_id_map.append_path(
                new_dir_path + "/" + os.path.basename(dir_path), target_id)
        except:
            pass

    def _try_load_terminal_auth_id(self):

        try:
            if not self._conf_path:
                return None

            with open(self._conf_path) as conf_file:
                flets_conf = json.load(conf_file)
                return flets_conf[self._terminal_key]
        except:
            return None

    def _save_terminal_auth_id(self, terminal_auth_id):

        try:
            with open(self._conf_path) as conf_file:
                flets_conf = json.load(conf_file)
        except:
            flets_conf = {}

        flets_conf[self._terminal_key] = terminal_auth_id

        with open(self._conf_path, "w") as conf_file:
            json.dump(flets_conf, conf_file)

    def _try_get_terminal_auth_id(self, headers):

        if not self._terminal_key:
            return None 

        terminal_auth_id = self._try_load_terminal_auth_id()
        if terminal_auth_id:
            return terminal_auth_id

        resp = self._request(
                method="POST",
                uri='terminal',
                headers=headers,
                body=urllib.urlencode({"terminal_id": self._terminal_id,
                                        "terminal_key": self._terminal_key}, True)
            )

        terminal_auth_id = resp["body"].findtext("terminal_auth_id", "")
        if terminal_auth_id:
            self._save_terminal_auth_id(terminal_auth_id)
            return terminal_auth_id
        return None

    def make_directory(self, dir_path):
        """Create "dir_path" directory."""

        if dir_path == "" or dir_path == FILE_SEPARATOR:
            # if blank, do nothing
            return

#        target_id = self._get_item_id(dir_path)
#        if target_id is not None:
#            # if exists, do nothing
#            return

        parent_path = os.path.dirname(dir_path)
        parent_id = self._get_item_id(parent_path, be_raise=True)

        target_name = os.path.basename(dir_path)

        if target_name == "" or target_name == FILE_SEPARATOR:
            # if blank, do nothing
            return

        try:
            self._request(
                method="POST",
                uri=["folder", parent_id],
                headers={"content-type": CONTENT_TYPE__URL_ENCODED},
                body=urllib.urlencode({"folder_name": target_name.encode("utf-8")})
            )
        except httplib2.HttpLib2Error, exc_ins:
            if str(exc_ins).find("error_code:21 ") >= 0:
                return

            if (str(exc_ins).find("error_code:9 ") >= 0 and
                self._stat(target_path=dir_path)[stat.ST_MTIME] &
                    stat.S_IFDIR):
                return

            raise

        try:
            new_id = resp["body"].findtext("file_id")
            self.path_id_map.append_path(dir_path, new_id)
        except:
            pass

    def delete_directory(self, dir_path):
        """Delete "dir_path" directory."""

        self._delete_item(target_path=dir_path)

    def stat_directory(self, dir_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        return self._stat(target_path=dir_path)

    def rename_directory(self, dir_path, new_name):
        """Rename a name of "dir_path" to "new_name"."""

        self._rename_item(target_path=dir_path, new_name=new_name)

    def move_directory(self, dir_path, new_dir_path):
        """Move "dir_path" to "new_path"."""

        self._move_item(target_path=dir_path,
                        new_dir_path=new_dir_path)

    def get_contents(self, dir_path, with_stat=False):
        """Return a contents dictionary in "dir_path".
           A stat information must be included, if "with_stat" is True,
        """

        items = self._get_item_list(
            parent_path=dir_path,
            parent_id=self._get_item_id(dir_path))

        contents = {}
        for item in items:
            item_path = os_path_bottom_name(item.findtext("file_path"))
            if with_stat:
                contents[item_path] = self._stat(target_item=item)
            else:
                contents[item_path] = {}

        return contents

    def get_updated_contents(self, dir_path):
        """Return a updated contents dictionary in "dir_path"."""

        raise NotImplementedError

    def open(self, file_path, mode="r", length=-1, chunk_size=None):
        """Return a class instance of the "FileObj" or sub class
           corresponding to "file_path".
        """

        self._debug_out("######## open %s\n" % (file_path,))
        return FletsAzukeruFileObj(self, file_path, mode, length, chunk_size)

    def delete_file(self, file_path):
        """Delete "file_path" file."""

        self._delete_item(target_path=file_path)

    def stat_file(self, file_path):
        """Return a stat information dictionary for "file_path"
           which keys below:
            stat.ST_MODE, stat.ST_MTIME, stat.CTIME, stat.ST_SIZE
        """

        return self._stat(target_path=file_path)

    def rename_file(self, file_path, new_name):
        """Rename a name of "file_path" to "new_name"."""

        try:
            parent_path = os.path.dirname(file_path)
            self.delete_file(parent_path + "/" + new_name)
        except:
            pass
        self._rename_item(target_path=file_path, new_name=new_name)

    def move_file(self, file_path, new_dir_path):
        """Move "file_path" to "new_path"."""

        self._move_item(target_path=file_path,
                        new_dir_path=new_dir_path)

    def test(self):
        pass
